# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    hoist.py                                           :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 11:16:28 by cacharle          #+#    #+#              #
#    Updated: 2021/02/11 20:13:29 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import re

import c_formatter_42.formatters.helper as helper


@helper.locally_scoped
def hoist(content: str) -> str:
    r""" Hoist local variable and split assigned declaration

    Assignment splitting:
    {                   {
        int a = 1;  =>      int a;
                            a = 1;
    }                   }

    Variable hoisting:
    {                         {
        puts("bonjour");          int a;
        int a;            =>      char b;
        char b;                   puts("bonjour");
    }                         }

    Only one empty line after declarations
    {                         {
                                  int a;
        int a;                    char b;
        puts("bonjour");  ->
                                  puts("bonjour");
        char b;               }
    }
    """

    input_lines = content.split("\n")

    lines = []
    # split assignment
    for line in input_lines:
        m = re.match(
            r"^(?P<indent>\s+)"
            r"(?P<type>{t})\s+"
            r"(?P<name>{d})\s+=\s+"
            r"(?P<value>.+);$"
            .format(t=helper.REGEX_TYPE, d=helper.REGEX_DECL_NAME),
            line
        )
        if m is not None:
            lines.append("\t{}\t{};".format(
                m.group("type"),
                m.group("name"))
            )
            lines.append("{}{} = {};".format(
                m.group("indent"),
                m.group("name").replace("*", ""),
                m.group("value"))
            )
        else:
            lines.append(line)

    # hoist declarations and filter empty lines
    decl_regex = r"^\s*{t}\s+{d};$".format(t=helper.REGEX_TYPE, d=helper.REGEX_DECL_NAME)
    declarations = [line for line in lines
                    if re.match(decl_regex, line) is not None]
    body = [line for line in lines
            if line not in declarations and line != ""]
    lines = declarations
    if len(declarations) != 0:
        lines.append("")
    lines.extend(body)

    return "\n".join(lines)



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    preprocessor_directive.py                          :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: charles <me@cacharle.xyz>                  +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2021/02/07 15:40:37 by charles           #+#    #+#              #
#    Updated: 2021/02/07 16:21:25 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import re


def preprocessor_directive(content: str) -> str:
    lines = content.split("\n")

    directive_regex = r"^\#\s*(?P<name>[a-z]+)\s?(?P<rest>.*)$"
    matches = [re.match(directive_regex, line) for line in lines]
    idented = [(i, match.group("name"), match.group("rest"))
               for i, match in enumerate(matches)
               if match is not None]
    indent = 0
    for i, directive_name, directive_rest in idented:
        if directive_name in ["elif", "else", "endif"]:
            indent -= 1
        lines[i] = "#" + (" " * indent) + directive_name + " " + directive_rest
        lines[i] = lines[i].strip()
        if directive_name in ["if", "ifdef", "ifndef", "elif", "else", "endif"]:
            indent += 1
        if directive_name == "endif":
            indent -= 1
    return "\n".join(lines)



# ############################################################################ #
#                                                                              #
#                                                         :::      ::::::::    #
#    clang_format.py                                    :+:      :+:    :+:    #
#                                                     +:+ +:+         +:+      #
#    By: cacharle <me@cacharle.xyz>                 +#+  +:+       +#+         #
#                                                 +#+#+#+#+#+   +#+            #
#    Created: 2020/10/04 10:40:07 by cacharle          #+#    #+#              #
#    Updated: 2021/02/12 11:38:45 by charles          ###   ########.fr        #
#                                                                              #
# ############################################################################ #

import os
import sys
import subprocess
from contextlib import contextmanager

import c_formatter_42

if sys.version_info >= (3, 7):
    from importlib.resources import path as resources_path
else:
    from importlib_resources import path as resources_path

CONFIG_FILENAME = ".clang-format"


@contextmanager
def _config_context():
    """ Temporarly place .clang-format config file in the current directory
        If there already is a config in the current directory, it's backed up
        then put back in place after clang-format is done running
    """
    with resources_path(c_formatter_42, CONFIG_FILENAME) as config_path:
        previous_config = None
        try:
            os.symlink(config_path, CONFIG_FILENAME)
        except FileExistsError:
            if not os.path.islink(CONFIG_FILENAME):
                with open(CONFIG_FILENAME) as f:
                    previous_config = f.read()
            os.unlink(CONFIG_FILENAME)
            os.symlink(config_path, CONFIG_FILENAME)
        yield
        os.unlink(CONFIG_FILENAME)
        if previous_config is not None:
            with open(CONFIG_FILENAME, "w") as f:
                f.write(previous_config)


def clang_format(content: str) -> str:
    """ Wrapper around the clang-format command

        Pass content on stdin and return stdout

        Note:
            Need to put .clang-format in user's home directory
            since clang-format doesn't allow to specify
            the path to a configuration file
    """
    with _config_context():
        process = subprocess.Popen(
            ["clang-format", "-style=file"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE
        )
        out, _ = process.communicate(input=content.encode())
        return out.decode()
